# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .node import ConfigNode
from ..namespace import Namespace, staticproperty
from ..utils import notnone_or
from .node_path import NodePath


class ComposedNode(ConfigNode):
    def __init__(self, children, nodes_memo=None, **kwargs):
        super().__init__(**kwargs)
        kwargs.pop('idx', None)
        kwargs.pop('metadata', None)
        kwargs.pop('delete', None)
        kwargs.pop('allow_new', None)
        kwargs.pop('safe', None)
        kwargs.update(self._get_child_kwargs())
        nodes_memo = nodes_memo if nodes_memo is not None else {}
        self._children = { name: ConfigNode(child, **kwargs, nodes_memo=nodes_memo) for name, child in children.items() } # pylint: disable=unexpected-keyword-arg

    class ayns(Namespace):
        def set_child(self, name, value):
            value = ConfigNode(value, **self._get_child_kwargs())
            self._children[name] = value
            value._propagate_implicit_values()
            return value

        def remove_child(self, name):
            child = self._children.pop(name, None)
            return child

        def rename_child(self, old_name, new_name):
            if not self.ayns.has_child(old_name):
                raise ValueError(f'{self!r} does not have a child named: {old_name!r}')
            if self.ayns.has_child(new_name):
                raise ValueError(f'Cannot rename a child named: {old_name!r} to {new_name!r}, the new name is already assigned to another child')

            child = self._children[old_name]
            del self._children[old_name]
            self._children[new_name] = child
            return child

        def get_child(self, name, default=None):
            return self._children.get(name, default)

        def has_child(self, name):
            if '_children' not in self.__dict__:
                return False
            return name in self._children

        @staticmethod
        def _get_node(root, access_fn, query_fn, *path, intermediate=False, names=False, incomplete=None):
            path = NodePath.get_list_path(*path)

            ret = []
            def _add(child, name):
                if names:
                    path = []
                    if ret:
                        path = ret[-1][2] + [name]
                    ret.append((child, name, path))
                else:
                    ret.append(child)

            _add(root, None)
            node = root
            found = True
            for component in path:
                if not query_fn(node, component): #not isinstance(node, ComposedNode) or not node.ayns.has_child(component):
                    _add(None, component)
                    found = False
                    break

                #node = node.ayns.get_child(component)
                node = access_fn(node, component)
                _add(node, component)

            if not found and not incomplete:
                if incomplete is None:
                    return None

                raise KeyError(f'{NodePath.join_path(path)!r} does not exist within {root!r}')

            return ret if intermediate else ret[-1]

        @staticmethod
        def _remove_node(root, remove_fn, *path):
            nodes = root.ayns.get_node(*path, intermediate=True, names=True, incomplete=None)
            if nodes is None:
                return None
            if len(nodes) == 1:
                raise ValueError(f'Cannot remove self from self')
            assert len(nodes) >= 2
            node, name, _ = nodes[-1]
            assert node is not None
            parent, _, _ = nodes[-2]
            return remove_fn(parent, name)

        def get_node(self, *path, intermediate=False, names=False, incomplete=False):
            ''' Inputs:
                    - `path` either a list of names (str or int) which should be looked up, or a str equivalent to calling NodePath.join_path on an analogical list
                    - `intermediate` if True, returned is a list of nodes accessed (including self), in order of accessing, otherwise only the final node is returned (i.e. the last element of the list)
                    - `names` if True, returned are tuples of `(node, name, path)`, where `node` is an object represting a node with name `name` (relative to its parent) and path `path` (list of names, relative to the self),
                        otherwise only the node object is returned.
                        This option applies to both cases: with and without `intermediate` set to True - in both cases either the single value returned, or all the values in the returned list, are either tuples of single objects.
                        If `intermediate` is True, or `path` is empty list/None, `self` can be also returned in which case `name` is None and `path` is an empty list.
                    - `incomplete` if evaluates to True and a node with path `path` cannot be found, returns the first missing node as `None` (if `names` is False) or `(None, name, path)` - in case `intermediate` is also set to True,
                        the last element of the returned list will be the first missing node with the rest of the list populated as usual. If `incomplete` is `None` and the node doesn't exist, `None` is returned (`intermediate` and `names` are ignored in that case),
                        otherwise a `KeyError` is raised.

                Returns:
                    - a node found by following path `path`, optionally together with its name and the full path (if `names` is True), if the node exists,
                    - otherwise if `incomplete` is set to True, a node "found" is the first missing node and is returned as `None` (again, optionally with name and a full path),
                    - regardless of whether a node was found or None was used to indicated a missing node, if `intermediate` is set to True returned is a list of all nodes
                        which were accessed (in order of accessing) before the final node was reached - this includes `self`, also each node can be returned either on its own or as a tuple together with their names and paths

                Raises:
                    `KeyError` if node cannot be found and `incomplete` evaluates to False and is not None
            '''
            def access_fn(node, component):
                return node.ayns.get_child(component)
            def query_fn(node, component):
                if not isinstance(node, ComposedNode):
                    return False
                return node.ayns.has_child(component)

            return ComposedNode.ayns._get_node(self, access_fn, query_fn, *path, intermediate=intermediate, names=names, incomplete=incomplete)

        def get_first_not_missing_node(self, *path, intermediate=False, names=False):
            def _get_node(n):
                return n if not names else n[0]
            nodes = self.ayns.get_node(*path, intermediate=True, names=names, incomplete=True)
            assert len(nodes) >= 2 or nodes[-1] is self
            if _get_node(nodes[-1]) is not None:
                return nodes[-1] if not intermediate else nodes

            nodes.pop()
            assert _get_node(nodes[-1]) is not None
            return nodes[-1] if not intermediate else nodes

        def remove_node(self, *path):
            def remove_fn(node, component):
                return node.ayns.remove_child(component)

            return ComposedNode.ayns._remove_node(self, remove_fn, *path)

        def replace_node(self, new_node, *path):
            raise NotImplementedError()

        def filter_nodes(self, condition, prefix=None, removed=None):
            prefix = NodePath.get_list_path(prefix, check_types=False) or NodePath()
            to_del = []
            to_re_set = []
            for name, child in self.ayns.named_children():
                child_path = prefix + [name]
                keep = False
                if condition(child_path, child):
                    keep = True
                if isinstance(child, ComposedNode): #not child.ayns.is_leaf:
                    possibly_new_child = child.ayns.filter_nodes(condition, prefix=child_path, removed=removed)
                    keep = keep or bool(possibly_new_child)
                    if keep and possibly_new_child is not child:
                        to_re_set.append(name, possibly_new_child)

                if not keep:
                    to_del.append(name)

            for name in reversed(to_del):
                self.ayns.remove_child(name)
                if removed is not None:
                    removed.add(prefix + [name])
            for name, child in to_re_set:
                self.ayns.set_child(name, child)

            return self

        def map_nodes(self, map_fn, prefix=None, cache_results=True, cache=None, leafs_only=True, include_self=True, recurse=True):
            prefix = NodePath.get_list_path(prefix, check_types=False) or NodePath()
            to_re_set = []
            if cache_results and cache is None:
                cache = {}

            for name, child in self.ayns.named_children():
                if cache_results and id(child) in cache:
                    possibly_new_child = cache[id(child)]
                else:
                    child_path = prefix + [name]
                    possibly_new_child = child
                    if isinstance(child, ComposedNode) and recurse:
                        possibly_new_child = possibly_new_child.ayns.map_nodes(map_fn, prefix=child_path, cache_results=cache_results, cache=cache, leafs_only=leafs_only)
                    else:
                        possibly_new_child = map_fn(child_path, possibly_new_child)
                        
                if possibly_new_child is not child:
                    to_re_set.append((name, possibly_new_child))

                if cache_results:
                    from ..utils import persistent_id
                    cache[persistent_id(child)] = possibly_new_child

            for name, child in to_re_set:
                self.ayns.set_child(name, child)

            ret = self
            if not leafs_only and include_self:
                ret = map_fn(prefix, self)

            return ret

        def nodes_with_paths(self, prefix=None, recursive=True, include_self=False, allow_duplicates=True):
            memo = set()
            prefix = NodePath.get_list_path(prefix, check_types=False) or NodePath()
            if include_self:
                memo.add(id(self))
                yield prefix, self

            for name, child in self._children.items():
                if child is None or (id(child) in memo and not allow_duplicates):
                    continue
                child_path = prefix + [name]
                if not recursive or not isinstance(child, ComposedNode):
                    memo.add(id(child))
                    yield child_path, child
                else:
                    for path, node in child.ayns.nodes_with_paths(prefix=child_path, recursive=recursive, include_self=True):
                        if id(node) in memo and not allow_duplicates:
                            continue
                        memo.add(id(node))
                        yield path, node

        def nodes(self, recursive=True, include_self=False, allow_duplicates=False):
            for _, node in self.ayns.nodes_with_paths(recursive=recursive, include_self=include_self, allow_duplicates=allow_duplicates):
                yield node

        def nodes_paths(self, prefix=None, recursive=True, include_self=False):
            for path, _ in self.ayns.nodes_with_paths(recursive=recursive, include_self=include_self, allow_duplicates=True):
                yield path


        def named_children(self, allow_duplicates=True):
            memo = set()
            for name, child in self._children.items():
                if not allow_duplicates:
                    if id(child) in memo:
                        continue
                    memo.add(id(child))
                yield name, child

        def children(self, allow_duplicates=True):
            for _, child in self.ayns.named_children(allow_duplicates=allow_duplicates):
                yield child

        def children_names(self):
            for key in self._children.keys():
                yield key

        def children_count(self, allow_duplicates=True):
            if allow_duplicates:
                return len(self._children)
            else:
                return len(set(self._children.values()))

        def clear(self):
            self._children.clear()

        def on_preprocess_impl(self, path, builder):
            return self.ayns.map_nodes(lambda child_path, node: node.ayns.on_preprocess(child_path, builder), prefix=path, cache_results=True, leafs_only=False, include_self=False, recurse=False)

        def on_premerge_impl(self, path, into):
            return self.ayns.map_nodes(lambda child_path, node: node.ayns.on_premerge(child_path, into), prefix=path, cache_results=True, leafs_only=False, include_self=False, recurse=False)

        def on_merge_impl(self, path, other):
            if not isinstance(other, ComposedNode):
                return ConfigNode.ayns.on_merge_impl(self, path, other)

            if other.ayns.delete:
                removed = set()
                def maybe_keep(path, node):
                    other_node = other.ayns.get_first_not_missing_node(path)
                    return node.ayns.has_priority_over(other_node)

                self.ayns.filter_nodes(maybe_keep, prefix=path, removed=removed)
                if not self._children and other.ayns.has_priority_over(self, if_equal=True):
                    removed.add(path)
                    other.ayns._require_all_new(path, f'note: the entire config tree under {path!r} has been removed due to node merging with a !del or !clear node', exceptions=removed)
                    ret = other._replace_other(self, allow_promotions=True)
                    return ret

            _this_path = NodePath.get_str_path(path)
            if not _this_path:
                _this_path = '<top-level node>'

            for key, value in other._children.items():
                child = self.ayns.get_child(key, None)
                if child is None:
                    value.ayns._require_all_new(path + [key], f'last parent: {_this_path!r}, from file: {self.ayns.source_file!r}')
                    self.ayns.set_child(key, value)
                else:
                    merge = isinstance(child, ComposedNode)
                    possibly_new_child = child.ayns.on_merge(path + [key], value)

                    if merge:
                        if not possibly_new_child and not possibly_new_child.ayns.has_priority_over(value) and value.ayns.explicit_delete:
                            self.ayns.remove_child(key)
                        elif possibly_new_child is not child:
                            self.ayns.set_child(key, possibly_new_child)
                    else:
                        if possibly_new_child is not child:
                            possibly_new_child.ayns._require_all_new(path + [key], f'last parent: {_this_path!r}, from file: {self.ayns.source_file!r}', include_self=False)
                            if not possibly_new_child and possibly_new_child.ayns.explicit_delete:
                                self.ayns.remove_child(key)
                            else:
                                self.ayns.set_child(key, possibly_new_child)

            if other.ayns.has_priority_over(self, if_equal=True):
                ret = self._replace_self(other, allow_promotions=True)
            else:
                ret = self._replace_other(other, allow_promotions=True)

            return ret

        @staticproperty
        @staticmethod
        def is_leaf():
            return False

        def _require_all_new(self, path, reason, exceptions=None, include_self=True):
            seq = self.ayns.nodes_with_paths(prefix=path, include_self=include_self)
            for p, n in seq:
                if not n.ayns.allow_new and (exceptions is None or p not in exceptions):
                    raise ValueError(f'Node {p!r} (source file: {n.ayns.source_file!r}) requires that the destination already exists but the current config tree does not contain a node under this path ({reason})')


    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_children']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    @staticmethod
    def _recreate(cls):
        new = cls.__new__(cls)
        new._children = {}
        return new

    def __reduce__(self):
        state = self.__getstate__()
        lit = None
        dit = None
        if isinstance(self, list):
            lit = iter(self)
        elif isinstance(self, dict):
            dit = iter(self.items())
        return ComposedNode._recreate, (type(self), ), state, lit, dit

    def _get_child_kwargs(self, child=None):
        ret = {}
        if not hasattr(self, '_delete'): # happens when unpickling! children are being populated before attributes are set, but its ok since we assume pickled objects are ok anyway, so no need to fix things
            return ret
        ret['implicit_delete'] = notnone_or(self._delete, self._default_delete or self._implicit_delete)
        ret['implicit_allow_new'] = notnone_or(self._allow_new, self._implicit_allow_new)
        if child is None or getattr(child, '_implicit_safe') is not False: # do not set "implicit_safe" arg if the child exists and already has it set to False (note: I think it's not strictly necessary to handle it here since other checks would still prevent changes)
            ret['implicit_safe'] = notnone_or(self._safe, self._implicit_safe)
        return ret

    def _propagate_implicit_values(self):
        if not hasattr(self, '_delete'): # happens when unpickling! children are being populated before attributes are set, but its ok since we assume pickled objects are ok anyway, so no need to fix things
            return
        if self._implicit_delete is None and self._implicit_allow_new is None and self._implicit_safe is None:
            return
        if self._delete is not None and self._allow_new is not None and self._safe is not None:
            return

        for child in self._children.values():
            fix = False
            if self._delete is None:
                if child._implicit_delete != self._implicit_delete:
                    child._implicit_delete = self._implicit_delete
                    fix = True
            if self._allow_new is None:
                if child._implicit_allow_new != self._implicit_allow_new:
                    child._implicit_allow_new = self._implicit_allow_new
                    fix = True
            if self._safe is None:
                if child._implicit_safe != self._implicit_safe:
                    if child._implicit_safe is not False:
                        child._implicit_safe = self._implicit_safe
                        fix = True

            if fix:
                child._propagate_implicit_values()

    @classmethod
    def _is_composed(cls):
        return True

    @classmethod
    def _is_plain_composed(cls):
        from .dict import ConfigDict
        from .list import ConfigList
        return cls in [ConfigList, ConfigDict]



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .nodes.node import ConfigNode
from .nodes.node_path import NodePath
from .namespace import NamespaceableMeta
from .utils import Bunch
from . import errors
from . import utils

import copy
import contextlib


class EvalContext(metaclass=NamespaceableMeta):
    ''' A class used to provide evaluation context for a top-level `awesomeyaml.nodes.ConfigDict`.
        After the top level node is put within a context, it can be evaluated to create `awesomeyaml.utils.Bunch`
        which will hold the evaluation result - evaluated nodes are no longer `awesomeyaml` nodes and can represent anything.
        The primary use-case for this class and the resulting `Bunch` object is to construct a `awesomeyaml.Config` object
        which basically combines the returned `Bunch` object with the original `ConfigDict`.
    '''

    class PartialChild(Bunch):
        def __init__(self, path, eval_ctx, cfgobj):
            self._path = path
            self._eval_ctx = eval_ctx
            self._cfgobj = cfgobj
            super().__init__({})

        def __getitem__(self, key):
            if key not in self:
                node = self._cfgobj[key]
                return self._eval_ctx.evaluate_node(node, self._path + [key])

            return super().__getitem__(key)

        def __getattr__(self, name):
            #if name not in self:
            #    raise AttributeError(f'Object {type(self).__name__!r} does not have attribute {name!r}')
            return self[name]

        def __setattr__(self, name, value):
            if name.startswith('_'):
                return super().__setattr__(name, value)

            if name in self.__dict__:
                raise ValueError('Name conflict!')

            self[name] = value

        def __delattr__(self, name):
            try:
                super().__delattr__(name)
            except AttributeError:
                del self[name]

        def get_or_set(self, key):
            return self.setdefault(key, EvalContext.PartialChild(self._path + [key], self._eval_ctx, self._cfgobj[key]))

    _default_eval_symbols = {}

    def __init__(self, config_dict, eval_symbols=None):
        ''' Arguments:
                config_dict : a `awesomeyaml.nodes.ConfigDict` object to be evaluated,
                    to construct a single `ConfigDict` from multiple sources
                    `awesomeyaml.builder.Builder` can be used.
                eval_symbols : a dict containing symbols which can be used when evaluating
                    ``config_dict``. The values from this argument will be used to update
                    the defaults from :py:meth:`get_default_eval_symbols`.
        '''
        self.cfg = config_dict
        self._removed_nodes = {}
        self._eval_cache = {}
        self._eval_cache_id = {}
        self._eval_symbols = copy.copy(EvalContext._default_eval_symbols)
        if eval_symbols:
            self._eval_symbols.update(eval_symbols)

        self._require_all_safe = False
        self._eval_stack = []

    @contextlib.contextmanager
    def require_all_safe(self, node, path):
        old, self._require_all_safe = self._require_all_safe, True
        sp = len(self._eval_stack)-1
        try:
            yield
        except errors.UnsafeError as e:
            raise errors.EvalError(f'The node requires all dependencies to be safe but the above error was raised indicating at least one !unsafe node was attempted to be executed while following the chain of dependencies: {self._eval_stack[sp:]}', node, path) from e
        finally:
            self._require_all_safe = old

    def get_node(self, *path, **kwargs):
        path = NodePath.get_list_path(*path)
        if str(path) in self._eval_cache:
            return self._eval_cache[str(path)]
        return self.cfg.ayns.get_node(path, **kwargs)

    def get_evaluated_node(self, nodepath):
        nodepath = NodePath.get_list_path(nodepath, check_types=False) or NodePath()
        node = self.ayns.get_node(nodepath)
        if str(nodepath) in self._eval_cache:
            return node

        return self.evaluate_node(node, prefix=nodepath)

    @errors.api_entry
    def evaluate_node(self, cfgobj, prefix=None):
        if not isinstance(cfgobj, ConfigNode):
            return cfgobj

        self._eval_stack.append(prefix)
        prefix = NodePath.get_list_path(prefix, check_types=False) or NodePath()

        if self._require_all_safe:
            if not cfgobj.ayns.safe:
                raise errors.UnsafeError(f'Note: the current context requires all evaluated nodes to be safe - see chained exceptions for more information', cfgobj, str(prefix))

        if id(cfgobj) in self._eval_cache_id:
            return self._eval_cache_id[id(cfgobj)]

        evaluated_parent = None
        if prefix:
            enode = self.ecfg
            for p in prefix[:-1]:
                assert isinstance(enode, EvalContext.PartialChild)
                enode = enode.get_or_set(p)

            evaluated_parent = enode

        evaluated_cfgobj = cfgobj.ayns.on_evaluate(prefix, self)
        if evaluated_parent is not None:
            evaluated_parent[prefix[-1]] = evaluated_cfgobj

        self._eval_cache[str(prefix)] = evaluated_cfgobj
        self._eval_cache_id[utils.persistent_id(cfgobj)] = evaluated_cfgobj
        self._eval_stack.pop()
        return evaluated_cfgobj

    @errors.api_entry
    def evaluate(self):
        ''' Returns:
                `awesomeyaml.utils.Bunch` representing evaluated config node.
        '''
        self._eval_cache.clear()
        self._eval_cache_id.clear()
        self.ecfg = EvalContext.PartialChild(NodePath(), self, self.cfg)
        ret = self.evaluate_node(self.cfg)
        self.ecfg = ret
        self._eval_cache.clear()
        self._eval_cache_id.clear()
        return ret

    @staticmethod
    def set_default_eval_symbols(symbols):
        ''' Sets the default symbols available to use when evaluating nodes.

            The symbols provided can be used by certain node types on evaluation.
            The defaults can be updated with values passed to the :py:meth:`__init__`
            method.

            Arguments:
                symbols : a lookup table (dict) of available symbols - for exampe, can be provided by
                    calling ``globals()`` in a module whose content one wishes to be available
                    during evaluation.

            Raises:
                TypeError: if ``symbols`` is not a dict.
        '''
        if not isinstance(symbols, dict):
            raise TypeError('dict expected')
        EvalContext._default_eval_symbols = symbols

    @staticmethod
    def get_default_eval_symbols():
        ''' Returns a dict with the default symbols available to use when evaluating nodes.

            The symbols provided can be used by certain node types on evaluation.
            The defaults can be updated with values passed to the :py:meth:`__init__`
            method.

            Returns:
                dict: symbols available by default
        '''
        return EvalContext._default_eval_symbols

    def get_eval_symbols(self):
        ''' Returns a dict with symbols available when evaluating nodes within this :py:class:`EvalContext`.

            Some node types might use this to evaluate their values.

            The returned value considers both the default symbols (see :py:meth:`set_default_eval_symbols`) and
            non-default ones passed to :py:meth:`__init__`.

            Returns:
                dict: symbols available in this evaluation context
        '''
        return self._eval_symbols



import re
import collections.abc as cabc


class NodePath(list):
    _path_component_regex = re.compile(r'''
                # the available options are:
                    (?:^|(?<=\.)) # ...if it is either at the beginning or preceded by a dot (do not capture)
                    ( [a-zA-Z0-9_]+ ) # a textual identifier (captured by the fist capture group)
                | # OR...
                    \[ (-?[0-9]+) \] # a (possibly negative) integer in [] (with second capture group catching the integer inside)
                | # OR...
                    (?<!^) # ... if it's not the first character ...
                    \. # a single dot
                    (?:(?!$)(?!\.)(?!\[)) # ... does not appear at the end and is not followed by either another dot or [ (do not capture)
                ''', re.VERBOSE) # verbose flag enables us to have comments, whitespace (inc. multi-line) etc. for better readability

    def __str__(self):
        return self.get_str_path(self)

    def __repr__(self):
        return repr(self.__str__())

    def __add__(self, other):
        return NodePath(list.__add__(self, other))

    def __hash__(self):
        return hash(str(self))

    @classmethod
    def split_path(cls, path_str, validate=True):
        matches = cls._path_component_regex.finditer(path_str)
        if validate:
            _matches_cache = []
            beg = None
            end = None
            for match in matches:
                _matches_cache.append(match)
                if beg is None:
                    beg = match.start()
                # in the first iteration, check is beg (i.e. match.start()) == 0 to make sure we do not have an invalid prefix
                # in every other iteration make sure the new match strictly follow the previous one, if it doesn't that means
                # that there was an invalid part between them (invalid here means not matching our regular expression)
                if match.start() != (end if end is not None else 0):
                    raise ValueError(f'Invalid path: {path_str!r}')
                end = match.end()

            # the path is fine at the beginning and internally doesn't have any "gaps" but has an incorrect suffix
            # (i.e. a part at the end which doesn't match our regular expression)
            if path_str and end != len(path_str):
                raise ValueError(f'Invalid path: {path_str!r}')

            matches = _matches_cache
        
        for match in matches:
            if match.group(1): # name
                yield str(match.group(1))
            elif match.group(2):  # index
                yield int(match.group(2))
            else:
                # sanity check
                assert match.group(0) == '.'

    @classmethod
    def join_path(cls, path_list):
        ret = ''
        for component in path_list:
            ret = ret + cls._get_child_accessor(component, ret)
        return ret

    @classmethod
    def _get_child_accessor(cls, childname, myname=''):
        if isinstance(childname, int):
            return f'[{childname}]'
        return ('.' if str(myname) else '') + str(childname)

    @classmethod
    def get_list_path(cls, *path, check_types=True):
        if not path:
            return NodePath()
        if len(path) == 1:
            if not isinstance(path[0], int):
                path = path[0]

        if path is None:
            return NodePath()
        if not isinstance(path, cabc.Sequence) or isinstance(path, str):
            # split_path should only return str and ints so we don't need to check for types
            path = cls.split_path(str(path))
        elif check_types:
            for i, c in enumerate(path):
                # we need to check it because if something is not a string nor an int it's ambiguous which casting should be done
                if not isinstance(c, str) and (not isinstance(c, int) or isinstance(c, bool)):
                    raise ValueError(f'node path should be a list of ints and/or str only, got {type(c)} at position {i}: {c}')

        return NodePath(path)

    @classmethod
    def get_str_path(cls, *path, check_types=True):
        if not path:
            return ''
        if len(path) == 1:
            path = path[0]

        if path is None:
            return ''
        if isinstance(path, int):
            return cls._get_child_accessor(path)
        if isinstance(path, cabc.Sequence) and not isinstance(path, str):
            return cls.join_path(path)

        if check_types and not isinstance(path, str):
            raise ValueError(f'Unexpected type: {type(path)}, expected int, sequence or str')
        return path



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import threading
import contextlib
import collections.abc as cabc

from .node_path import NodePath
from ..namespace import NamespaceableMeta, Namespace, staticproperty
from ..utils import persistent_id, notnone_or
from ..import errors


_kwargs_to_inherit = [
    'priority',
    'implicit_delete',
    'implicit_allow_new',
    'implicit_safe',
    'pyyaml_node'
]


def decorator_factory(error_type):
    def decorator(func):
        def impl(*args, **kwargs):
            self = args[0]
            path = args[1] if len(args) > 1 else kwargs['path']
            other = args[2] if len(args) > 2 else kwargs.get('other', None)
            if not isinstance(other, ConfigNode):
                other = None
            with errors.rethrow(error_type, self, path, other):
                return func(*args, **kwargs)

        return impl
    return decorator


rethrow_as_preprocess_error = decorator_factory(errors.PreprocessError)
rethrow_as_premerge_error = decorator_factory(errors.PremergeError)
rethrow_as_merge_error = decorator_factory(errors.MergeError)
rethrow_as_eval_error = decorator_factory(errors.EvalError)


class ConfigNodeMeta(NamespaceableMeta):
    def __call__(cls,
            *args,
            nodes_memo=None,
            _force_type=False,
            **kwargs):
        value = None
        has_value = False
        if args:
            value = args[0]
            args = args[1:]
            has_value = True

        # check if type deduction is required
        if cls is ConfigNode and not _force_type:
            # deduce type and call it recursively (this time enforcing it)
            if not has_value:
                raise ValueError('Cannot deduce target type without a positional argument - deduction is always done w.r.t. the first argument')
            if isinstance(value, ConfigNode):
                for arg_name in _kwargs_to_inherit:
                    if arg_name in kwargs:
                        # do not change implicit_safe if already set to False
                        if arg_name == 'implicit_safe' and getattr(value, '_' + arg_name) is False:
                            del kwargs[arg_name]
                            continue
                        setattr(value, '_' + arg_name, kwargs[arg_name])
                if any(k.startswith('implicit_') for k in kwargs.keys()):
                    value._propagate_implicit_values()

                return value
            else:
                from .dict import ConfigDict
                from .list import ConfigList
                from .tuple import ConfigTuple
                from .scalar import ConfigScalar

                if isinstance(value, cabc.Sequence) and not isinstance(value, str) and not isinstance(value, bytes):
                    if isinstance(value, cabc.MutableSequence):
                        t = ConfigList
                    else:
                        t = ConfigTuple
                elif isinstance(value, cabc.MutableMapping):
                    t = ConfigDict
                else:
                    t = ConfigScalar

            # dispatch actual object creation (see below)
            # we need to do that recursively since __call__ method can be overwritten (e.g. ConfigScalar)
            return t(value, *args, nodes_memo=nodes_memo, _force_type=True, **kwargs)

        # actual object creation
        if has_value and nodes_memo is not None and id(value) in nodes_memo:
            return nodes_memo[id(value)]

        from .composed import ComposedNode
        if issubclass(cls, ComposedNode):
            kwargs['nodes_memo'] = nodes_memo

        if has_value:
            ret = NamespaceableMeta.__call__(cls, value, *args, **kwargs)
        else:
            assert not args
            ret = NamespaceableMeta.__call__(cls, **kwargs)

        if has_value and nodes_memo is not None:
            assert id(value) not in nodes_memo
            nodes_memo[persistent_id(value)] = ret

        return ret


class ConfigNode(metaclass=ConfigNodeMeta):
    WEAK = -1
    STANDARD = 0
    FORCE = 1

    special_metadata_names = [
        'idx',
        'priority',
        'delete',
        'allow_new',
        'source_file',
        'safe'
    ]

    _default_filename = threading.local()
    _default_safe = threading.local()
    _default_priority = STANDARD
    _default_delete = False
    _default_allow_new = True

    @staticmethod
    @contextlib.contextmanager
    def default_filename(filename):
        if not hasattr(ConfigNode._default_filename, 'value'):
            ConfigNode._default_filename.value = None

        old = ConfigNode._default_filename.value
        ConfigNode._default_filename.value = filename
        try:
            yield
        finally:
            ConfigNode._default_filename.value = old

    @staticmethod
    @contextlib.contextmanager
    def default_safe_flag(value):
        if not hasattr(ConfigNode._default_safe, 'value'):
            ConfigNode._default_safe.value = True

        old = ConfigNode._default_safe.value
        ConfigNode._default_safe.value = value and old
        try:
            yield
        finally:
            ConfigNode._default_safe.value = old


    def __init__(self, idx=None, priority=None, delete=None, allow_new=None, safe=None, metadata=None, source_file=None, implicit_delete=None, implicit_allow_new=None, implicit_safe=None, pyyaml_node=None):
        ''' '''

        """ There's a lit bit going on with merging flags here, basically the main idea is that we have 3 sources of flags, they are (in the precedence order):
                 - explicit merging-controlling information attached to this node (highest precedence), this are passed as "delete", "allow_new", etc.
                 - merging flags inherited from a parent node (aka implicit flags) - these can only have not-None value if there's an ancestor node with explicit flag, however the immediate parent does not have to have a flag specified explicitly
                 - a node type's defaults
        """
        if priority not in [None, ConfigNode.STANDARD, ConfigNode.WEAK, ConfigNode.FORCE]:
            raise ValueError(f'Unknown priority value: {priority}')
        self._idx = idx
        self._priority = priority
        self._delete = delete
        self._allow_new = allow_new
        self._implicit_delete = implicit_delete
        self._implicit_allow_new = implicit_allow_new
        self._source_file = source_file if source_file is not None else getattr(ConfigNode._default_filename, 'value', None)
        self._metadata = metadata or {}
        self._pyyaml_node = pyyaml_node
        self._safe = safe
        self._implicit_safe = implicit_safe
        self._default_safe = getattr(ConfigNode._default_safe, 'value', False)

    def __repr__(self, simple=False):
        return f'<Object {type(self).__name__!r} at 0x{id(self):02x}>'

    class ayns(Namespace):
        @property
        def idx(self):
            return self._idx

        @property
        def priority(self):
            if self._priority is None:
                return self._default_priority

            return self._priority

        @property
        def weak(self):
            return self.ayns.priority == ConfigNode.WEAK

        @property
        def force(self):
            return self.ayns.priority == ConfigNode.FORCE

        @property
        def delete(self):
            if self._delete is None:
                if self._implicit_delete is not None:
                    return self._implicit_delete
                return self._default_delete

            return self._delete

        @property
        def allow_new(self):
            if self._implicit_allow_new is not None:
                return self._implicit_allow_new
            return self._default_allow_new

        @property
        def explicit_delete(self):
            return self._delete

        @property
        def source_file(self):
            return self._source_file

        @property
        def metadata(self):
            return self._metadata

        @property
        def safe(self):
            return notnone_or(self._safe, True) and notnone_or(self._implicit_safe, True) and notnone_or(self._default_safe, False)

        @staticproperty
        @staticmethod
        def is_leaf():
            return True

        @staticproperty
        @staticmethod
        def is_root():
            return False

        @property
        def value(self):
            return self._get_value()

        @property
        def native_value(self):
            return self._get_native_value()

        @value.setter
        def value(self, value):
            return self._set_value(value)

        @staticproperty
        @staticmethod
        def tag():
            return None

        def has_priority_over(self, other, if_equal=False):
            if self.ayns.priority == other.ayns.priority:
                return if_equal
            return self.ayns.priority > other.ayns.priority


        #
        # The following functions follow a high-level naming conention as follows:
        #   - <name> - is a top level method that should be called on the top-level config objects only
        #   - on_<name> - is an equivalent that is being called recursively for all nodes in the config object when <name> is called
        #   - on_<name>_impl - is the part of "on_<name>" that can/should be overwritten by derived classes - it should always be called indirectly via "on_<name>"
        #
        # Note: "evaluate" does not exist as it is implemented in EvalContext (since we need to keep state in a separate class) - we could have "evaluate" here
        # and delegate functionality to the EvalContext but to prevent possible errors it's better to remove it altogether (e.g., to avoid accidental recursive calls)

        def preprocess(self, builder):
            return self.ayns.on_preprocess(NodePath(), builder)

        @rethrow_as_preprocess_error
        def on_preprocess(self, path, builder):
            return self.ayns.on_preprocess_impl(path, builder)

        def on_preprocess_impl(self, path, builder):
            return self


        def premerge(self, into=None):
            return self.ayns.on_premerge(NodePath(), into)

        @rethrow_as_premerge_error
        def on_premerge(self, path, into):
            return self.ayns.on_premerge_impl(path, into)

        def on_premerge_impl(self, path, into):
            return self


        def merge(self, other):
            if other is None:
                if not self.ayns.allow_new:
                    raise ValueError('A top-level destination node does not exist but this top-level node has a !notnew flag enabled!')
                return self

            other.ayns.premerge(self)
            return self.ayns.on_merge(NodePath(), other)

        @rethrow_as_merge_error
        def on_merge(self, path, other):
            return self.ayns.on_merge_impl(path, other)

        def on_merge_impl(self, path, other):
            if self.ayns.has_priority_over(other):
                self._replace_other(other, allow_promotions=False)
                return self
            other._replace_other(self, allow_promotions=False)
            return other


        @rethrow_as_eval_error
        def on_evaluate(self, path, root):
            evaluated = self.ayns.on_evaluate_impl(path, root)
            assert evaluated is not self
            assert not isinstance(evaluated, ConfigNode)
            return evaluated

        def on_evaluate_impl(self, path, ctx):
            return self.ayns.value


        def get_node_info_to_save(self):
            ''' This function should return a dict with values which one wants to preserve when dumping the node.
            '''
            # by default we don't care about node idx or source file (after dumping this will change anyway)
            # we care about custom metadata and "priority" and "delete"
            # Note that "priority" and "delete" might be optimized out from the dump output if
            # they would be set by the parent (dumping function holds a stack of which metadata are "default"
            # and does not produce anything which is aligned with the defaults)
            ret = copy.copy(self._metadata)
            ret['priority'] = self._priority
            ret['delete'] = self._delete #if not self._implicit_delete else None
            ret['allow_new'] = self._allow_new
            ret['safe'] = self._safe
            return ret

        @property
        def node_info(self):
            return {
                'idx': self._idx,
                'priority': self._priority,
                'delete': self._delete,
                'implicit_delete': self._implicit_delete,
                'allow_new': self._allow_new,
                'implicit_allow_new': self._implicit_allow_new,
                'safe': self._safe,
                'implicit_safe': self._implicit_safe,
                'default_safe': self._default_safe,
                'source_file': self._source_file,
                'metadata': self._metadata
            }

        def get_default_mode(self):
            return {
                'priority': self._default_priority,
                'delete': self._default_delete,
                'allow_new': self._default_allow_new,
                'safe': self._default_safe
            }

        def represent(self):
            ''' Returns a tuple ``(tag, metadata, data)``, where ``tag`` is desired tag (can be ``None``),
                ``metadata`` is a dict with metadata (optional, can evaluate to ``False`` to ignore),
                and ``data`` is object which will be used to recursively represent ``self`` (can be either
                mapping, sequence or scalar).
            '''
            return self.ayns.tag, self.ayns.get_node_info_to_save(), self.ayns.value

        def _require_all_new(self, path, reason, exceptions=None, include_self=True):
            if not include_self:
                return
            if not self.ayns.allow_new and (exceptions is None or path not in exceptions):
                raise ValueError(f'Node {path!r} (source file: {self._source_file!r}) requires that the destination already exists but the current config tree does not contain a node under this path ({reason})')

        def _require_safe(self, path):
            if not self.ayns.safe:
                raise errors.UnsafeError(None, self, path)

    def _propagate_implicit_values(self):
        return

    #
    # The following methods are used to implement some core merging semantics - they are intended to be called whenever two nodes are merged.
    # Depending on the desired outcome, if we are merging nodes "A <- B", one of the four possible outcomes should happen:
    #   - if B has higher priority than A and we want to use B's memory location to store the output, one should call "B._replace_other(A)"
    #   - if B has higher priority than A but we want to use A's memory location to store the output, one should call "A._replace_self(B)"
    #   - if B has lower priority and we want to use B to store the result, "B._replace_self(A)" should be called
    #   - finally, if B has lower priority and A is to be used to hold the result, "A._replace_other(B)" should be called
    #
    # In summary, the generic rule: "<resulting_node_obj>._replace_{self|other}(<other_node_obj>)" where <resulting_node_obj> is the object
    # that is intended to be returned by the merging method and "{self|other}" depends on the relative priority between the two nodes.
    #
    # These methods >> DO NOT << handle the logic to perform merging of the nodes' content, it is assumed that the "<resulting_node_obj>" already
    # holds correct content - which is the result of the merge - at the moment of calling these methods.
    # However, the methods might still modify some class-specific content via promotion mechanism (conditioned on the relevant argument).
    # Promotion happen when merging "A <- B" and one of the nodes has more specific type than the other, e.g., A is "!bind" node
    # and B is a dict with "!del" - in such a case, it can happen that B is going to be the result of the merge (if the entire subtree under node A
    # is deleted), but that would make the resulting type to be "dict", rather than "!bind". To ensure correct behaviour, promotion mechanism will
    # attempt to promote the "!bind" node to be the result rather than "dict" by moving the entire content of the dict node into the "!bind" and
    # returning the latter. Therefore the final result is equivalent to returning "dict" but with extra elements specific to the "!bind" node.
    #

    def _replace_self(self, other, allow_promotions=False):
        ''' Helper that should be called whenever "other" is merged into "self" with higher priority,
            but we want to keep the resulting object inside the memory of "self".

            Intuitively speaking, it will try to make "self" look like "other".
        '''
        self._priority = other._priority
        self._delete = other._delete
        if other._safe is not None:
            self._safe = notnone_or(self._safe, True) and other._safe
        if other._default_safe is not None:
            self._default_safe = notnone_or(other._default_safe, True) and other._default_safe
        self._metadata = { **self._metadata, **other._metadata }
        if allow_promotions:
            ret = self._maybe_promote(other)
        else:
            ret = self
        ret._propagate_implicit_values()
        return ret

    def _replace_other(self, other, allow_promotions=False):
        ''' Helper that should be called whenever "other" is merged into "self" with lower priority,
            and we want to keep "self" as the resulting node (subject to promotions).

            Intuitively speaking, it will selectively add content of "other" into "self" to make sure that
            any extra information is preserved, while keeping the content already present in "self" as it is.

            If "allow_promotions" is set to True, "other" can be returned instead of "self" if its type is preferred.
            In such case, its content will be identical to the content of "self" after replacement has been done,
            with any extra content coming from other's type preserved.
        '''
        if other._safe is not None:
            self._safe = notnone_or(self._safe, True) and other._safe
        if other._default_safe is not None:
            self._default_safe = notnone_or(other._default_safe, True) and other._default_safe
        self._metadata = { **other._metadata, **self._metadata }
        if allow_promotions:
            return self._maybe_promote(other)
        return self

    def _maybe_promote(self, other):
        ''' Possibly promote "other" to be returned rather than "self" if its type is preferred.

            This function >> DOES NOT << raise any errors related to types, i.e., it assumes all checks have already
            been made and a correct merge is FINISHING, with the last step being promoting types, if needed. Consequently,
            if promotion is not possible, the method WILL NOT REPORT any errors, instead will simply give up.

            Argument convention is that "self" is about to replace "other" as the result of merging.

            If "other" is promoted, it should replicate "self" as much as possible.
        '''
        if type(self) is type(other):
            return self
        if not self._is_composed() or not other._is_composed():
            return self

        if issubclass(type(other), type(self)): # simple case, plain dict/list replaces complex dict/list, promote complex
            other.clear()
            if isinstance(other, list):
                other.extend(self)
            else:
                other.update(self)
            other.__dict__.update(self.__dict__)
            return other
        elif issubclass(type(self), type(other)): # complex dict/list replaces simple dict/list, leave as is
            return self
        elif self._is_plain_composed() and not other._is_plain_composed():  # plain dict/list replaces complex list/dict, promote complex
            other.clear()
            if isinstance(other, list):
                other.extend(self.values())
            else:
                other.update(enumerate(self))
            other.__dict__.update(self.__dict__)
            return other
        elif not self._is_plain_composed() and other._is_plain_composed(): # complex dict/list replaces simple list/dict, leave as is
            return self
        else: # any other case, silently give up
            return self

    @classmethod
    def _is_composed(cls):
        return False

    @classmethod
    def _is_plain_composed(cls):
        return False



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .scalar import ConfigScalar
from .node import ConfigNode
from ..namespace import namespace, staticproperty
from ..utils import Bunch, python_is_at_least
from ..errors import EvalError

import os
import dis
import sys
import types
import hashlib


class GlobalsWrapper():
    def __init__(self, gbls, ecfg, ctx, node, path):
        self.gbls = gbls
        self.ecfg = ecfg
        self.ctx = ctx
        self.node = node
        self.path = path

    def __getattr__(self, name):
        if name in self.gbls:
            return self.gbls[name]

        if name in self.ecfg._cfgobj:
            with self.ctx.require_all_safe(self.node, self.path):
                return self.ecfg[name]
        elif name in __builtins__:
            return __builtins__[name]
        else:
            raise NameError(name)

class EvalNode(ConfigScalar(str)):
    ''' Implements ``!eval`` tag.

        The eval node can be used to evaluate arbitrary python code.
        Internally, it is treated exactly as normal string node with the
        only exception being that on evaluation its value is passed to ``eval``.

        More specifically, the value obtained from evaluating an eval node is the
        value returned by the last line of code store in the node. If more lines are
        present, the preceding ones are executed before the last one to provide
        additional context (e.g., by defining functions, importing modules etc.).

        On top of that, the code stored in an eval node can use symbols defined
        by a relevant :py:class:`EvalContext` (please see its constructor)
        and can access all top-level nodes in the config by their names.

        The summary of the eval node's behaviour is illustrated by the following snippet::

            symbols = eval_context.get_eval_symbols()
            symbols.update(eval_context.get_top_level_objects_and_names())
            lines = str(self.code).split('\\n')
            exec('\\n'.join(lines[:-1]), symbols)
            return eval(lines[-1], symbols)

        Supported syntax::

            !eval code

        Merge behaviour:

            The same as standard string node.
    '''
    _top_namespace_module_name = 'awesomeyaml.eval_node_namespace'
    _globals_wrapper_name = '__ayns_globals_wrapper'

    def __init__(self, value, persistent_namespace=True, **kwargs):
        super().__init__(value, **kwargs)
        self.persistent_namespace = persistent_namespace

    @namespace('ayns')
    def on_evaluate_impl(self, path, ctx):
        self.ayns._require_safe(path)
        code_hash = hashlib.md5(str(self).encode('utf-8')).hexdigest()
        eval_module_name = f'{EvalNode._top_namespace_module_name}.{str(path).replace(".", "_")}_0x{code_hash}'

        from_module = False
        if self.persistent_namespace and eval_module_name in sys.modules:
            gbls = sys.modules[eval_module_name].__dict__
            from_module = True
        else:
            gbls = {
                'ayns': Bunch({
                    'ctx': ctx,
                    'cfg': ctx.ecfg
                })
            }
            gbls.update(ctx.get_eval_symbols())
            gbls.update({ '__name__': eval_module_name, '__file__': self._source_file })

        gbls[EvalNode._globals_wrapper_name] = GlobalsWrapper(gbls, ctx.ecfg, ctx, self, path)

        lines = self.strip().split('\n')
        lines = [lline for line in lines for lline in line.split(';')]

        exec_lines = "\n".join(lines[:-1])
        eval_line = lines[-1].strip()

        try:
            exec_code = compile(exec_lines, self._source_file, 'exec')
            eval_code = compile(eval_line, self._source_file, 'eval')
            exec_code_patched, _ = EvalNode._patch_access_to_globals(exec_code)
            eval_code_patched, _ = EvalNode._patch_access_to_globals(eval_code)
            exec(exec_code_patched, gbls)
            ret = eval(eval_code_patched, gbls)
        except EvalError as e:
            code = f'=== CODE BEGINS ===\n{os.linesep.join(lines)}\n=== CODE ENDS ==='
            if e.node is self:
                e.note = code
                raise
            else:
                raise EvalError('The above exception occurred in the user code.', self, path, note=code) from e
        except Exception as e:
            code = f'=== CODE BEGINS ===\n{os.linesep.join(lines)}\n=== CODE ENDS ==='
            raise EvalError('The above exception occurred in the user code.', self, path, note=code) from e

        del gbls[EvalNode._globals_wrapper_name]

        if len(lines) > 1 and self.persistent_namespace and not from_module:
            eval_node_module = types.ModuleType(eval_module_name, 'Dynamic module to evaluate awesomeyaml !eval node')
            eval_node_module.__dict__.update(gbls)
            sys.modules[eval_module_name] = eval_node_module

        if isinstance(ret, ConfigNode):
            assert ret is not self
            ret = ctx.evaluate_node(ret, path)
        return ret

    @namespace('ayns')
    @staticproperty
    @staticmethod
    def tag():
        return '!eval'


    @staticmethod
    def _patch_access_to_globals(code):
        done_something = False

        def maybe_patch(const):
            if isinstance(const, types.CodeType):
                nonlocal done_something
                new_const, done_something_sub = EvalNode._patch_access_to_globals(const)
                if done_something_sub:
                    done_something = True
                return new_const

            return const

        new_consts = [maybe_patch(const) for const in code.co_consts]
        new_names = list(code.co_names)
        has_wrapper = bool(EvalNode._globals_wrapper_name in new_names)

        _wrapper_idx = None
        def get_wrapper_index():
            nonlocal _wrapper_idx
            if _wrapper_idx is None:
                if has_wrapper:
                    _wrapper_idx = new_names.index(EvalNode._globals_wrapper_name)
                else:
                    new_names.append(EvalNode._globals_wrapper_name)
                    _wrapper_idx = len(new_names) - 1

            assert _wrapper_idx is not None
            return _wrapper_idx


        new_bytecode = []
        rjumps = []
        ajumps = []
        location_map = {}

        i = 0
        while i < len(code.co_code):
            done = False
            op = code.co_code[i]
            if dis.opname[op] in ['LOAD_GLOBAL', 'LOAD_NAME']:
                arg = code.co_code[i+1]
                arg_shifted = (python_is_at_least(3, 11) and dis.opname[op] == 'LOAD_GLOBAL')

                if arg_shifted:
                    null_bit = arg & 0x01
                    arg = arg>>1

                name = code.co_names[arg]

                if name not in ['__ayns_globals_wrapper', 'ayns']:
                    new_arg = get_wrapper_index()
                    if arg_shifted:
                        new_arg = (new_arg << 1) | null_bit

                    location_map[i//2] = len(new_bytecode)
                    new_bytecode.append(code.co_code[i:i+1] + new_arg.to_bytes(1, 'little'))
                    if python_is_at_least(3, 11):
                        caches = dis._inline_cache_entries[op]
                        for _ in range(caches):
                            assert i+2 < len(code.co_code)
                            assert code.co_code[i+2] == 0
                            new_bytecode.append(code.co_code[i+2:i+4])
                            i += 2

                    new_bytecode.append(dis.opmap['LOAD_ATTR'].to_bytes(1, 'little') + arg.to_bytes(1, 'little'))
                    if python_is_at_least(3, 11):
                        caches = dis._inline_cache_entries[dis.opmap['LOAD_ATTR']]
                        for _ in range(caches):
                            new_bytecode.append(b'\x00\x00') # CACHE(0)

                    done_something = True
                    done = True
            elif op in dis.hasjabs:
                ajumps.append(len(new_bytecode))
            elif op in dis.hasjrel:
                rjumps.append(len(new_bytecode))

            if not done:
                location_map[i//2] = len(new_bytecode)
                new_bytecode.append(code.co_code[i:i+2])

            i += 2

        if not done_something:
            return code, False

        rlocation_map = { value: key for key, value in location_map.items() }

        for rjump in rjumps:
            new_jump_loc = rjump
            old_jump_loc = rlocation_map[new_jump_loc]

            op, old_loc_rel = new_bytecode[rjump]
            is_backward = False
            if python_is_at_least(3, 11) and 'BACKWARD' in dis.opname[op]:
                is_backward = True

            if not python_is_at_least(3, 10): # python 3.10 starts counting instructions (each 2-bytes long), before that it was counting bytes
                # in this code the convention is to count instructions (e.g., because of the way "new_bytecode" is organized), so translate
                # offset from bytes to instructions
                old_loc_rel //= 2

            if is_backward:
                old_loc_abs = old_jump_loc - old_loc_rel
            else:
                old_loc_abs = old_loc_rel + old_jump_loc

            new_loc_abs = location_map[old_loc_abs]
            new_loc_rel = abs(new_loc_abs - new_jump_loc)
            if not python_is_at_least(3, 10):
                new_loc_rel *= 2

            new_bytecode[rjump] = op.to_bytes(1, 'little') + new_loc_rel.to_bytes(1, 'little')

        for ajump in ajumps:
            op, old_loc_abs = new_bytecode[ajump]
            if not python_is_at_least(3, 10):
                old_loc_abs //= 2
            new_loc_abs = location_map[old_loc_abs]
            if not python_is_at_least(3, 10):
                new_loc_abs *= 2
            new_bytecode[ajump] = op.to_bytes(1, 'little') + new_loc_abs.to_bytes(1, 'little')


        new_bytecode = b''.join(new_bytecode)
        if python_is_at_least(3, 11):
            return types.CodeType(code.co_argcount,
                                    code.co_posonlyargcount, # only in Python >= 3.8
                                    code.co_kwonlyargcount,
                                    code.co_nlocals,
                                    code.co_stacksize,
                                    code.co_flags,
                                    new_bytecode,
                                    tuple(new_consts),
                                    tuple(new_names),
                                    code.co_varnames,
                                    code.co_filename,
                                    code.co_name,
                                    code.co_qualname, # only in Python >= 3.11
                                    code.co_firstlineno,
                                    code.co_lnotab,
                                    code.co_exceptiontable, # only in Python >= 3.11
                                    code.co_freevars,
                                    code.co_cellvars), True
        elif python_is_at_least(3, 8):
            return types.CodeType(code.co_argcount,
                                    code.co_posonlyargcount, # only in Python >= 3.8
                                    code.co_kwonlyargcount,
                                    code.co_nlocals,
                                    code.co_stacksize,
                                    code.co_flags,
                                    new_bytecode,
                                    tuple(new_consts),
                                    tuple(new_names),
                                    code.co_varnames,
                                    code.co_filename,
                                    code.co_name,
                                    code.co_firstlineno,
                                    code.co_lnotab,
                                    code.co_freevars,
                                    code.co_cellvars), True
        else:
            return types.CodeType(code.co_argcount,
                                    code.co_kwonlyargcount,
                                    code.co_nlocals,
                                    code.co_stacksize,
                                    code.co_flags,
                                    new_bytecode,
                                    tuple(new_consts),
                                    tuple(new_names),
                                    code.co_varnames,
                                    code.co_filename,
                                    code.co_name,
                                    code.co_firstlineno,
                                    code.co_lnotab,
                                    code.co_freevars,
                                    code.co_cellvars), True



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import threading
import contextlib
import yaml.error


rethrow = True # whether errors happening inside awesomeyaml should be rethrown as higher-level errors (listed below)
include_original_exception = True # if exceptions are rethrown, whether to include the original exception as "__cause__" of the new exception
shorten_traceback = True # if exceptions are rethrown, whether to shorten their stack by removing some intermediate frames

_api_entered = threading.local()


def api_entry(fn):
    def impl(*args, **kwargs):
        if getattr(_api_entered, 'value', False) or not rethrow or not shorten_traceback:
            return fn(*args, **kwargs)

        _api_entered.value = True
        try:
            return fn(*args, **kwargs)
        except Error as e:
            if include_original_exception:
                orig_exp = e.__context__
                if orig_exp is not None:
                    orig_exp.__traceback__ = orig_exp.__traceback__.tb_next # skip "rethrow_as_XXX"
                reason = orig_exp
            else:
                reason = None

            raise type(e)(
                error_msg=e.error_msg,
                node=e.node,
                path=e.path,
                extra_node=e.extra_node,
                note=e.note
            ) from reason
        finally:
            _api_entered.value = False

    return impl


def _get_mark_or_fallback_str(aynode):
    if aynode is None:
        return None
    if aynode._pyyaml_node is not None:
        return aynode._pyyaml_node.start_mark
    return f'<A node from file: {aynode._source_file!r}>'


class Error(yaml.error.MarkedYAMLError):
    def __init__(self, error_msg, node, path=None, extra_node=None, note=None):
        self.error_msg = error_msg
        self.node = node
        self.path = path
        self.extra_node = extra_node

        msg = self.base_msg

        try:
            types = repr(type(self.node).__name__)
            if self.extra_node is not None:
                types += f' and {type(self.extra_node).__name__!r}'

            path_str = repr(path if path else '<top-level node>') + f' from file {self.node._source_file!r}'
            if self.extra_node is not None:
                path_str += f' and {self.extra_node._source_file!r}'

            msg = msg.format(types, path_str)
        except:
            pass

        if error_msg:
            if msg:
                msg += '\n'
            msg += error_msg

        if self.extra_node is None:
            if self.stage == 'parsing':
                if self.node is not None:
                    mark = self.node.start_mark
                else:
                    mark = None
            else:
                mark = _get_mark_or_fallback_str(self.node)

            super().__init__(
                problem=msg,
                problem_mark=mark,
                note=note
            )
        else:
            super().__init__(
                context=msg,
                context_mark=_get_mark_or_fallback_str(self.node),
                problem_mark=_get_mark_or_fallback_str(self.extra_node),
                note=note
            )


class ParsingError(Error):
    stage = 'parsing'
    base_msg = 'Parsing error occurred'


class PreprocessError(Error):
    stage = 'preprocess'
    base_msg = 'An error occurred while preprocessing a {} node under path: {}'

class PremergeError(Error):
    stage = 'premerge'
    base_msg = 'An error occurred while performing pre-merge actions for a {} node under path: {}'


class MergeError(Error):
    stage = 'merge'
    base_msg = 'An error occurred while merging {} nodes under path: {}'

class EvalError(Error):
    stage = 'eval'
    base_msg = 'An error occurred while evaluating a {} node under path: {}'


class UnsafeError(EvalError):
    base_msg = 'Avoiding execution of an !unsafe {} node under path: {}'


@contextlib.contextmanager
def rethrow(error_type, self, path, other):
    try:
        yield
    except error_type as e:
        if shorten_traceback:
            raise
        else:
            raise error_type(error_msg=None, node=self, path=path, extra_node=other) from e
    except Exception as e:
        if rethrow:
            reason = None
            if include_original_exception:
                reason = e

            raise error_type(error_msg=str(e), node=self, path=path, extra_node=other) from reason
        else:
            raise
